import os
import argparse
import itertools
from pywebcat.utils import WebCAT


def main():
    """Command line function for saving frames of webCAT videos.
    """
    (
        directory,
        station,
        year,
        month,
        day,
        time,
        interval,
        no_meta,
        quality,
        verbose,
    ) = parse_args()
    wc = WebCAT()
    for item in itertools.product(station, year, month, day, time):
        try:
            wc.generate_url(*item)  # generate url from the input data
            tmp_dir = os.path.join(directory, item[0], wc.name)  # dir to save frames in
            if not os.path.exists(tmp_dir):
                os.makedirs(tmp_dir)  # mkdir if not exist
            if verbose:
                print(f"Saving frames of {wc.name}...")
            wc.save_frames(
                interval, tmp_dir, not no_meta, quality, verbose
            )  # save frames
        except Exception as e:
            if verbose:
                url = f"http://webcat-video.axds.co/{item[0]}/raw/{item[1]}/{item[1]}_{item[2]:02}/{item[1]}_{item[2]:02}_{item[3]:02}/{item[0]}.{item[1]}-{item[2]:02}-{item[3]:02}_{item[4]:04}.mp4"
                print(f"Error for url: {url}. Traceback: {e} Skipping this url.")


def parse_args():
    parser = argparse.ArgumentParser(
        description="CLI for saving frames of webCAT video(s)."
    )
    required = parser.add_argument_group("required arguments")
    # below I'm using --arg with "required" to allow for multiple "nargs"
    required.add_argument(
        "-dir",
        "--directory",
        type=dir_path,
        help="Absolute path of directory to save frames in.",
        required=True,
    )
    required.add_argument(
        "-s",
        "--station",
        nargs="+",
        type=str,
        help="The station name, e.g., buxtoncoastalcam.",
        required=True,
    )
    required.add_argument(
        "-y",
        "--year",
        nargs="+",
        type=int,
        help="The video year(s), e.g., 2019 2020.",
        required=True,
    )
    required.add_argument(
        "-m",
        "--month",
        nargs="+",
        type=int,
        help="The video month(s), e.g., 9 10 11.",
        required=True,
    )
    required.add_argument(
        "-d",
        "--day",
        nargs="+",
        type=int,
        help="The video day(s) e.g., 1 11 21.",
        required=True,
    )
    required.add_argument(
        "-t",
        "--time",
        nargs="+",
        type=int,
        help="The video time(s), e.g., 1000 1330 1510.",
        required=True,
    )
    parser.add_argument(
        "-i",
        "--interval",
        default=10,
        type=int,
        help="Interval in seconds between video frames to save (default: 10).",
    )
    parser.add_argument(
        "-n",
        "--no_meta",
        action="store_true",
        help="Don't save .csv file of metadata of saved video frames.",
    )
    parser.add_argument(
        "-q",
        "--quality",
        default=95,
        type=int,
        help="Quality of saved frames between 0 (lowest) and 100 (highest) (default: 95).",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Print program status."
    )
    args = parser.parse_args()

    return (
        args.directory,
        args.station,
        args.year,
        args.month,
        args.day,
        args.time,
        args.interval,
        args.no_meta,
        args.quality,
        args.verbose,
    )


def dir_path(path: str):
    """Utility function for argparse to check existence of a passed directory.

    Parameters
    ----------
    path : str
        Absolute path to local directory.

    """

    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(
            f"readable_dir:{path} is not a valid directory path."
        )


if __name__ == "__main__":
    main()



import os
import cv2
import shutil
import itertools
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import urllib.request
from tqdm import tqdm

plt.style.use("ggplot")
plt.rcParams.update(
    {
        "font.size": 14,
        "axes.labelweight": "bold",
        "figure.figsize": (8, 6),
        "axes.grid": False,
    }
)


class WebCAT:
    """Utility class for viewing local or remote webcat videos.

    Attributes
    ----------
    url :
        url to the .mp4 file.
    name :
        Unique name based on url.
    video : cv2.VideoCapture
        The cv2.VideoCapture object of the video file at url.
    frames: int
        Total number of frames in the video.
    fps : int
        Frames per second in the video.
    width : int
        Width of video frames in pixels.
    height : int
        Height of video frames in pixels.

    Methods
    -------
    download_url(self, fout=None, verbose=1)
        Download the video from the url.

    Examples
    --------
    >>> from webcat_utils import VidRetriever
    >>> url = "http://webcat-video.axds.co/buxtoncoastalcam/raw/2019/2019_11/2019_11_13/buxtoncoastalcam.2019-11-13_1000.mp4"
    >>> vr = VidRetriever(url)
    >>> vr.download_url()
    ...
    Saving to buxtoncoastalcam.2019-11-13_1000.mp4:  0%|      | 44.1M/125M [00:08<00:16, 5.18MB/s]
    """

    def __init__(self):
        self.url = None
        self.name = None
        self.video = None

    @property
    def width(self):
        return int(self.video.get(3))

    @property
    def height(self):
        return int(self.video.get(4))

    @property
    def frames(self):
        return int(self.video.get(7))

    @property
    def fps(self):
        return int(self.video.get(cv2.CAP_PROP_FPS))

    def generate_url(self, station: str, year: int, month: int, day: int, time: int):
        """Generate WebCAT URLs and expressive name for files from user inputs.

        Parameters
        ----------
        station : str
            Station name, e.g., "buxtoncoastalcam".
        year : int
            Year of video, e.g., 2020.
        month: int
            Month (numerical) of video, e.g., 11.
        day: int
            Day of video, e.g., 17.
        time: int
            Time (24 hr) of video rounded to nearest 10 minutes, e.g., 0500 (5:00 am), 1300 (1:00 pm), 1330 (1:30pm).

        """
        url = f"http://webcat-video.axds.co/{station}/raw/{year}/{year}_{month:02}/{year}_{month:02}_{day:02}/{station}.{year}-{month:02}-{day:02}_{time:04}.mp4"
        try:
            urllib.request.urlopen(url)
        except:
            raise ValueError(f"{url} is not a valid url.")
        vid = cv2.VideoCapture(url)
        self.url = url
        self.video = vid
        self.name = f"{station}_{year}_{month}_{day}_{time}"

    def download_url(self, fout: str = None, verbose: bool = True):
        """Download the video from the instance url.

        Parameters
        ----------
        fout : str
            The file path to save the downloaded file to, e.g., ~/Downloads/download.mp4, by default None.
        verbose : {True, False}, optional
            Display download progress bar, by default True.
        """
        fout = self.name + ".mp4" if fout == None else fout
        if verbose:
            with TqdmUpTo(
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                miniters=1,
                desc=f"Saving to {fout}",
            ) as t:  # all optional kwargs
                urllib.request.urlretrieve(self.url, fout, reporthook=t.update_to)
        else:
            urllib.request.urlretrieve(self.url, fout)

    def save_frames(
        self,
        delta_t: int = 10,
        fout_path: str = "",
        save_csv=True,
        quality: int = 95,
        verbose=False,
    ):
        """Download the video from the instance url.

        Parameters
        ----------
        delta_t : int, optional
            A frame will be saved every delta_t seconds, by default 10.
        fout_path : str, optional
            Path to save frames and csv to, e.g., "~/Downloads/".
        save_csv : {True, False}, optional
            Save a csv fiel containing saved frame metadata.
        quality: int, optional
            Quality of the written JPEG images between 0 (lowest quality) and 100 (highest quality), by default 95.
        verbose : {True, False}, optional
            Display progress bar, by default True.

        Examples
        --------
        >>> from webcat_utils import WebCAT
        >>> wc.generate_url("buxtoncoastalcam", 2019, 11, 13, 1000)
        >>> wc.save_frames(delta_t=100, fout_path="./", save_csv=False)
        """
        if delta_t >= int(self.frames / self.fps):
            raise ValueError(
                f"delta_t should be less than {int(self.frames / self.fps)}."
            )
        step = delta_t * self.fps
        step_range = range(0, (self.frames), step)
        loop = tqdm(step_range) if verbose else step_range
        tmp_dir = os.path.join(fout_path, "jpg")  # save images in a "jpg" folder
        if os.path.exists(tmp_dir):
            shutil.rmtree(tmp_dir)  # just remove the whole directory
        os.makedirs(tmp_dir)  # mkdir
        for i in loop:
            self.video.set(1, i)
            _, frame_arr = self.video.read()
            cv2.imwrite(
                os.path.join(tmp_dir, f"frame_{i}.jpg"),
                frame_arr,
                [cv2.IMWRITE_JPEG_QUALITY, quality],
            )
        if save_csv:
            (
                pd.DataFrame(
                    {
                        "url": self.url,
                        "name": self.name,
                        "frame": list(step_range),
                        "path": [
                            os.path.join(tmp_dir, f"frame_{_}.jpg") for _ in step_range
                        ],
                    }
                ).to_csv(os.path.join(fout_path, f"{self.name}.csv"))
            )

    def plot_frames(self, frames: list = [0]):
        """Plot frames from the video, zero-based indexing.

        Parameters
        ----------
        frames : list, optional
            List of the frames to display in a grid plot, by default [0].

        Returns
        -------
        numpy.ndarray of matplotlib.AxesSubplot objects

        Examples
        --------
        >>> from webcat_utils import WebCAT
        >>> wc.generate_url("buxtoncoastalcam", 2019, 11, 13, 1000)
        >>> wc.plot_frames()
        ...
        array([[<matplotlib.axes._subplots.AxesSubplot object at 0x12c8922d0>]],
              dtype=object)
        """
        if not all((_ >= 0) & (_ < self.frames) for _ in frames):
            raise ValueError(f"frames should be between 0 and {self.frames-1}.")
        rows = 1 + (len(frames) - 1) // 3
        cols = len(frames) if len(frames) < 3 else 3
        fig, ax = plt.subplots(rows, cols, figsize=(cols * 4, rows * 3), squeeze=False)
        for i, frame in enumerate(frames):
            plt.sca(ax[0, i]) if rows == 1 else plt.sca(ax[i // 3, i % 3])
            self.video.set(1, frame)
            _, frame_arr = self.video.read()
            plt.imshow(
                cv2.cvtColor(frame_arr, cv2.COLOR_BGR2RGB)
            )  # plot the image on current axis
            plt.title(f"Frame {frame + 1} of {self.frames}")
        if len(frames) < rows * cols:  # clear any unsued axes (this is dirty)
            for j in np.arange(len(frames), rows * cols):
                ax[j // 3, j % 3].set_visible(False)
        plt.tight_layout()

        return ax

    def plot_average_frame(self, step: int = 500):
        """Plot the average of every "step" frames in the video.

        Parameters
        ----------
        step : int, optional
            The step between frames to average by, lower values result in a smoother average, by default 10.

        Returns
        -------
        matplotlib.AxesSubplot

        Examples
        --------
        >>> from webcat_utils import WebCAT
        >>> wc.generate_url("buxtoncoastalcam", 2019, 11, 13, 1000)
        >>> wc.plot_average_frame()
        ...
        <matplotlib.axes._subplots.AxesSubplot>
        """
        N = self.frames // step  # how many frames to average based on step
        timex = np.zeros((self.height, self.width, 3, N), np.float)  # init array
        for i in range(N):  # I dislike loops, could probably make this much faster
            self.video.set(1, i * step)
            _, frame_arr = self.video.read()
            timex[:, :, :, i] = cv2.cvtColor(frame_arr, cv2.COLOR_BGR2RGB)
        timex = np.mean(timex, axis=-1).astype(int)
        fig, ax = plt.subplots(1, 1, figsize=(10, 7))
        plt.imshow(timex)
        plt.title(f"Average of every {step} frames, {N} frames in total")

        return ax


class TqdmUpTo(tqdm):
    """Utility class for displaying a progress bar.

    Methods
    -------
    download_url(self, fout=None, verbose=1)
        Download the video from the url.

    Examples
    --------
    >>> url = "http://webcat-video.axds.co/buxtoncoastalcam/raw/2019/2019_11/2019_11_13/buxtoncoastalcam.2019-11-13_1000.mp4"
    >>> with TqdmUpTo(unit="B", unit_scale=True, unit_divisor=1024, miniters=1, desc=f"Saving to {fout}") as t:
    >>>     urllib.request.urlretrieve(self.fin, fout, reporthook=t.update_to)
    ...
    Saving to buxtoncoastalcam.2019-11-13_1000.mp4:  0%|      | 44.1M/125M [00:08<00:16, 5.18MB/s]
    """

    def update_to(self, b: int = 1, bsize: int = 1, tsize: int = None):
        """Progress bar udpate function.

        Parameters
        ----------
        b : int, optional
            Number of blocks transferred so far, by default 1
        bsize : int, optional
             Size of each block (in tqdm units), by default 1
        tsize : int, optional
            Total size (in tqdm units), by default None
        """

        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pywebcat']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18.5,<2.0.0',
 'opencv-python>=4.2.0,<5.0.0',
 'pandas>=1.0.4,<2.0.0',
 'tqdm>=4.46.1,<5.0.0']

entry_points = \
{'console_scripts': ['pywebcat = pywebcat.cli:main']}

setup_kwargs = {
    'name': 'pywebcat',
    'version': '0.2.0',
    'description': 'Python tool for working with the NOAA NOS Web Camera Applications Testbed (WebCAT)',
    'long_description': '# PyWebCAT\n\n[![PyPI version](https://badge.fury.io/py/pywebcat.svg)](https://badge.fury.io/py/pywebcat)\n[![Documentation Status](https://readthedocs.org/projects/pywebcat/badge/?version=latest)](https://pywebcat.readthedocs.io/en/latest/?badge=latest)\n![build](https://github.com/UNCG-DAISY/PyWebCAT/workflows/build/badge.svg?branch=master)\n[![codecov](https://codecov.io/gh/UNCG-DAISY/PyWebCAT/branch/master/graph/badge.svg)](https://codecov.io/gh/UNCG-DAISY/PyWebCAT)\n[![DOI](https://zenodo.org/badge/268162661.svg)](https://zenodo.org/badge/latestdoi/268162661)\n\nThis repository provides a Pythonic way to interface with the NOAA National Ocean Service Web Camera Applications \nTestbed (WebCAT). The real-time data is hosted on the [SECOORA site](https://secoora.org) \n(The Southeast Coastal Ocean Observing Regional Association), on the dedicated [WebCAT page](https://secoora.org/webcat/).\nHistoric data can also be accessed by retrieving files using specific HTTP requests (using a pattern described on the \n[WebCAT page](https://secoora.org/webcat/)). \n\nMore details about WebCAT are available in this Open Access paper:\n\n>Dusek, G., Hernandez, D., Willis, M., Brown, J. A., Long, J. W., Porter, D. E., & Vance, T. C. (2019). WebCAT: Piloting the development of a web camera coastal observing network for diverse applications. Frontiers in Marine Science, 6, 353, 25 June 2019 | https://doi.org/10.3389/fmars.2019.00353\n\n## Install\n\n```{sh}\npip install pywebcat\n```\n\n## Usage\n\n### CLI\n\nA key workflow for utilising WebCAT videos is to split videos into frames for further analysis. As a result, this functionality is exposed as a CLI with the command line argument `pywebcat`.\n\n```{sh}\n$ pywebcat --help\nusage: webcat_utils.py [-h] -dir DIRECTORY -s STATION [STATION ...] -y YEAR\n                       [YEAR ...] -m MONTH [MONTH ...] -d DAY [DAY ...] -t\n                       TIME [TIME ...] [-i INTERVAL] [-n] [-v]\n\nCLI for saving frames of webCAT video(s).\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -i INTERVAL, --interval INTERVAL\n                        Interval in seconds between video frames to save\n                        (default: 10).\n  -n, --no_meta         Don\'t save .csv file of metadata of saved video\n                        frames.\n  -q QUALITY, --quality QUALITY\n                        Quality of saved frames between 0 (lowest) and 100\n                        (highest) (default: 95).\n  -v, --verbose         Print program status.\n\nrequired arguments:\n  -dir DIRECTORY, --directory DIRECTORY\n                        Absolute path of directory to save frames in.\n  -s STATION [STATION ...], --station STATION [STATION ...]\n                        The station name, e.g., buxtoncoastalcam.\n  -y YEAR [YEAR ...], --year YEAR [YEAR ...]\n                        The video year(s), e.g., 2019 2020.\n  -m MONTH [MONTH ...], --month MONTH [MONTH ...]\n                        The video month(s), e.g., 9 10 11.\n  -d DAY [DAY ...], --day DAY [DAY ...]\n                        The video day(s) e.g., 1 11 21.\n  -t TIME [TIME ...], --time TIME [TIME ...]\n                        The video time(s), e.g., 1000 1330 1510.\n```\n\nThe CLI facilitates efficiently looping over input arguments to locate one or more WebCAT videos and split them into a desired number of frames which are then saved locally. Here\'s an example looping over several videos from the Buxton coastal camera on 13th Nov 2019 at different times (a 10 second interval is specified for saving the frames and verbosity is turned on):\n\n```{sh}\n$ pywebcat -dir /Users/tbeuzen/Downloads -s buxtoncoastalcam -y 2019 -m 11 -d 13 -t 1000 1210 1530 -i 10 -v\nSaving frames of buxtoncoastalcam_2019_11_13_1000...\n100%|██████████████████████████████████████████████| 62/62 [00:14<00:00,  4.22it/s]\nSaving frames of buxtoncoastalcam_2019_11_13_1210...\n100%|██████████████████████████████████████████████| 62/62 [00:13<00:00,  4.51it/s]\nSaving frames of buxtoncoastalcam_2019_11_13_1530...\n100%|██████████████████████████████████████████████| 62/62 [00:14<00:00,  4.38it/s]\n```\n\nThe resultant directory structure looks like:\n\n```{sh}\nUsers/tbeuzen/Downloads\n                └── buxtoncoastalcam\n                    ├── buxtoncoastalcam_2019_11_13_1000\n                    │   ├── buxtoncoastalcam_2019_11_13_1000.csv\n                    │   └── jpg\n                    │       ├── frame_0.jpg\n                    │       ├── frame_280.jpg\n                    │       ├── ...\n                    ├── buxtoncoastalcam_2019_11_13_1210\n                    │   ├── ...\n                    └── buxtoncoastalcam_2019_11_13_1530\n                        ├── ...\n```\n\nThe outputted .csv file contains metadata for the saved frames:\n\n| url                                                                                                           | name                             | frame | path                                                                                         |\n|---------------------------------------------------------------------------------------------------------------|----------------------------------|-------|----------------------------------------------------------------------------------------------|\n| http://webcat-video.axds.co/buxtoncoastalcam/raw/2019/2019_11/2019_11_13/buxtoncoastalcam.2019-11-13_1000.mp4 | buxtoncoastalcam_2019_11_13_1000 | 0     | /Users/tbeuzen/Downloads/buxtoncoastalcam/buxtoncoastalcam_2019_11_13_1000/jpg/frame_0.jpg   |\n| http://webcat-video.axds.co/buxtoncoastalcam/raw/2019/2019_11/2019_11_13/buxtoncoastalcam.2019-11-13_1000.mp4 | buxtoncoastalcam_2019_11_13_1000 | 280   | /Users/tbeuzen/Downloads/buxtoncoastalcam/buxtoncoastalcam_2019_11_13_1000/jpg/frame_280.jpg |\n| ...                                                                                                           | ...                              | ...   | ...                                                                                          |\n\n### Module\n\nThe pywebcat utilities can also be imported through the `pywebcat.utils` for use in other libraries or workflows. See the [demo Jupyter notebook](notebooks/pywebcat_demo.ipynb) for a worked example.\n\n```{python}\nfrom pywebcat.utils import WebCAT\nwc = WebCAT()\nwc.generate_url("buxtoncoastalcam", 2019, 11, 13, 1000)  # create the video url\n\n# attributes\nwc.url     # the created url\nwc.name    # unique name for the video object\nwc.width   # frame width in pixels\nwc.height  # frame height in pixels\nwc.frames  # total frames in video\nwc.fps     # frames per second\n\n# methods\nwc.download_url()        # download the video at the url\nwc.save_frames()         # save video frames as .jpg\nwc.plot_frames()         # plot select video frames\nwc.plot_average_frame()  # plot time-averaged frame\n```\n\n## Contributing\n\nContributions are welcome and greatly appreciated! If you\'re interested in contributing to this project, take a look at the [contributor guide](docs/contributing.rst).\n\n## Contributors\n\nAll contributions are welcomed and recognized! You can see a list of current contributors in the [contributors tab](https://github.com/UNCG-DAISY/pywebcat/graphs/contributors).\n\n## Acknowledgements\n\nThanks to the [Southeast Coastal Ocean Observing Regional Association (SECOORA) and the WebCAT project](https://secoora.org/webcat/) for making camera footage publicly available and supporting open science and data.\n',
    'author': 'Tomas Beuzen',
    'author_email': 'tomas.beuzen@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/UNCG-DAISY/WebCAT-Utilities',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
